/*
* Module Name:     aipsocdkl.c
*
* Description:     Cash Drawer & KeyLock Driver for Linux on SureOne 4614.
*
* (C) Copyright 2005, International Business Machines Corporation.
* All right reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* 1.  Redistributions of source code must retain the above copyright notice,
*     this list of conditions and the following disclaimer.
*
* 2.  Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*
* 3.  The name of the author may not be used to endorse or promote products
*     derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* Should you need to contact me, the author, you can do so either by
* e-mail - mail your message to <nobody@us.ibm.com>.
*/



/*
* Linux defs and headers
*/
#ifndef __KERNEL__
    #define __KERNEL__
#endif
#ifndef MODULE
    #define MODULE
#endif

#ifdef MODVERSIONS
    #include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/kernel.h>

/*
* Local Include - kernel/dcs/aipkernel.h
*/
#ifndef __UNIX__
    #define __UNIX__
#endif
#include "aipkernel.h"



/*
* SureOne machine specific declarations
*/
#define SUREONE 0xFE040                 /* Model Number Address in Bios */
#define SUREONEID 0x34313634            /* "4614" */

#define IBM_4613  8						/* Anyplace Kiosk ID as a command paramater of aipstart srcipt*/
#define IBM_4614  7                     /* SureOne ID as command paramater of aipstart script*/

/*
* Driver specific declarations
*/
#define DEV_NODE_NAME  "aipsocdkl"
#define AIPSOCDKL_MAJOR_DEV 249 

MODULE_AUTHOR("IBM Retail Store Solutions");
MODULE_LICENSE("GPL compatible");
MODULE_DESCRIPTION("Driver for the SureOne Cash Drawer & KeyLock");



static int machine_type = IBM_4614;  /*Default to IBM_4614*/
module_param(machine_type,int,0);
MODULE_PARM_DESC(machine_type,"Integer Value");

/*
*  ASIC Port definitions
*/
#define INDEX_PORT_ADDRESS     0x240        /* ASIC Data port index selector (W) */
#define DATA_PORT_ADDRESS      0x241        /* Data I/O Port Address */
#define KL_STATUS_INDEX        0x26         /* Keylock status data port (R) */
#define CD_SETUP_INDEX         0x28         /* Cash Drawer setup data port (RW) */
#define CD_OPEN_INDEX          0x29         /* Cash Drawer firing port (W) */
#define CD_STATUS_INDEX        0x30         /* Cash Drawer status port (RW) */

#define CD_ENABLE              0x06         /* xxxx x110   Use ASIC, 100ms,  /Enable  */
#define CD_OPEN                0x01         /* xxxx xxx1  Fire Cash Drawer */



/*
* Return Code
*/
#define SUCCESS 0



/*
* Debug Macros
*/
#ifdef DEBUG
    #define DPRINTK(fmt, args...) printk("%s(%s): " fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)
#else
    #define DPRINTK(fmt, args...)
#endif
#define EPRINTK(fmt, args...) printk(KERN_ERR "%s(%s): "     fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)
#define IPRINTK(fmt, args...) printk(KERN_INFO "%s(%s): "    fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)
#define WPRINTK(fmt, args...) printk(KERN_WARNING "%s(%s): " fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)



/*
* Function Prototypes
*/
static int  __init  aipsocdkl_init(void);
static void __exit  aipsocdkl_exit(void);
static int  aipsocdkl_open(struct inode *inode, struct file *flip);
static int  aipsocdkl_read(struct file *flip, char *buf, size_t count, loff_t *f_pos);
static int  aipsocdkl_write(struct file *flip, const char *buf, size_t count, loff_t *f_pos);
static int  aipsocdkl_release(struct inode *inode, struct file *flip);
static int  aipsocdkl_ioctl(struct inode *inode, struct file *flip, unsigned int cmd, unsigned long arg);
static int  aipsocdkl_ioctl_GetSlotInfo( char *pBuffer );
static int  aipsocdkl_ioctl_GetKLStatus( char *pBuffer );
static int  aipsocdkl_ioctl_OpenCD( char *pBuffer );
static int  aipsocdkl_ioctl_GetCDStatus( char *pBuffer );

static int aipsocdkl_major = AIPSOCDKL_MAJOR_DEV;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
MODULE_PARM(aipsocdkl_major, "i");
#else
module_param_named(major, aipsocdkl_major, int, 0644);
#endif

/*
* Device methods
*/
struct file_operations aipsocdkl_fops = {
    owner:    THIS_MODULE,
    read:     aipsocdkl_read,
    write:    aipsocdkl_write,
    ioctl:    aipsocdkl_ioctl,
    open:     aipsocdkl_open,
    release:  aipsocdkl_release,
};



/*
* Driver Global Variables
*/
static unsigned char * pBiosModelNumber;
static unsigned long CheckModel;



/************************************************************************
*
* Function Name:      aipsocdkl_init
*
* Purpose:            Intialization routine
*
* Description:        Entry point for this driver.
*
* Output:             0 - Success, Non-zero - Failure
*
* ***********************************************************************
*/
static int __init aipsocdkl_init(void)
{
    static unsigned short rc  = 0;

    pBiosModelNumber = (unsigned char *) ioremap(SUREONE, 4);
    CheckModel = readl(pBiosModelNumber);
    iounmap(pBiosModelNumber);

    /*
    * Don't install the driver, if it's not a SureOne or Anyplace Kiosk machine.
    */
    if (machine_type == IBM_4613) {
    } else if (machine_type == IBM_4614)
    {
        pBiosModelNumber = (unsigned char *) ioremap(SUREONE, 4);
        CheckModel = readl(pBiosModelNumber);
        iounmap(pBiosModelNumber);

        if (CheckModel != SUREONEID) {
          EPRINTK("This does not seem to be a SureOne or Anyplace Kiosk machine, model 0x%08x\n", (unsigned)  CheckModel);
          EPRINTK("This driver is for a SureOne or Anyplace Kiosk machines only\n");
          return -EINVAL;
       }
		IPRINTK( "found SureOne machine, model 0x%08x\n", (unsigned) CheckModel);
	}

    if ((rc = register_chrdev(aipsocdkl_major, DEV_NODE_NAME, &aipsocdkl_fops)))
    {
        EPRINTK("failed to register device with the kernel.  error %us\n",  rc);
        return -EINVAL;
    }

    DPRINTK("device registered with the kernel, major %i minor 0\n", aipsocdkl_major);

    /*
    * Enable the Cash Drawer, controll it through the ASIC and set default pulse width 100ms
    */
    outb(CD_SETUP_INDEX, INDEX_PORT_ADDRESS);
    outb(CD_ENABLE, DATA_PORT_ADDRESS);

    DPRINTK( "initialised cash drawer\n");

    return SUCCESS;
}



/************************************************************************
*
* Function Name:      aipsocdkl_exit
*
* Purpose:            Teardown
*
* Description:        Called when module is about to unload
*
* ***********************************************************************
*/
static void __exit aipsocdkl_exit(void)
{
    static unsigned short rc  = 0;

    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    if ( MOD_IN_USE )
    {
        WPRINTK( "device busy, remove delayed\n" );
        return;
    }
    #endif

    if ((rc = unregister_chrdev(aipsocdkl_major, DEV_NODE_NAME)))
    {
        EPRINTK( "unregister with kernel failed, error = %us\n", rc);
        return;
    }

    DPRINTK( "module removed\n");
}



/************************************************************************
*
* Function Name:      aipsocdkl_open
*
* Purpose:            Device Open
*
* Description:        Called by client to access driver services
*
* Output:             0 - Success, Non-zero - Failure
*
* ***********************************************************************
*/
static int aipsocdkl_open( struct inode *inode, struct file *flip )
{
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_INC_USE_COUNT;
    #endif
    DPRINTK("open\n");
    return SUCCESS;
}



/************************************************************************
*
* Function Name:      aipsocdkl_release
*
* Purpose:            Device Close
*
* Description:        Called by client to release access to driver services
*
* Output:             0 - Success
*
* ***********************************************************************
*/
static int aipsocdkl_release( struct inode *inode, struct file *flip )
{
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_DEC_USE_COUNT;
    #endif
    DPRINTK("close\n");
    return SUCCESS;
}



/************************************************************************
*
* Function Name:      aipsocdkl_read
*
* Notes:              Read Not supported
*
* ***********************************************************************
*/
static int aipsocdkl_read( struct file *flip,
                        char *buf,
                        size_t count,
                        loff_t *f_pos )
{
    DPRINTK( "read unsupported.\n");
    return -EINVAL;
}



/************************************************************************
*
* Function Name:      aipsocdkl_write
*
* Notes:              Write Not supported
*
* ***********************************************************************
*/
static int aipsocdkl_write( struct file *flip,
                            const char *buf,
                            size_t count,
                            loff_t *f_pos )
{
    DPRINTK( "write unsupported.\n");
    return -EINVAL;
}



/************************************************************************
*
* Function Name:      aipsocdkl_ioctl
*
* Purpose:            Request Device service
*
* Description:        Called by client to access device services
*
* Inputs:             Dependent on the IOCTL
*
* Output:             RPDONE
*                     0 - Success, -EINVAL - Unknown IOCTL
*                     Other Non-zero values - specific service errors
*
* ***********************************************************************
*/
static int aipsocdkl_ioctl( struct inode *inode,
                            struct file *flip,
                            unsigned int cmd,
                            unsigned long arg )
{
    switch (_IOC_NR(cmd))
    {

        case _IOC_NR(GET_KL_STATUS):
        {
            DPRINTK( "GET_KL_STATUS\n");
            return aipsocdkl_ioctl_GetKLStatus( (char *) arg );
            break;
        }

        case _IOC_NR(OPEN_CD):
        {
            DPRINTK( "OPEN_CD\n");
            return aipsocdkl_ioctl_OpenCD( (char *) arg );
            break;
        }

        case _IOC_NR(GET_CD_STATUS):
        {
            DPRINTK( "GET_CD_STATUS\n");
            return aipsocdkl_ioctl_GetCDStatus( (char *) arg );
            break;
        }

        case _IOC_NR(GET_SLOT_INFO):
        {
            DPRINTK( "GET_SLOT_INFO\n");
            return aipsocdkl_ioctl_GetSlotInfo( (char *) arg );
            break;
        }

        default:
        {
            DPRINTK( "Invalid IOCTL %u\n",cmd);
            return (-EINVAL);
            break;
        }
    }
}



/************************************************************************
*
* Function Name:      aipsocdkl_ioctl_GetKLStatus
*
* Purpose:            Get Keylock status
*
* Description:        The keylock can be in any of 3 positions. This fn.
*                     reads the port at index 0x26 and examines bits
*                     2..0.
*
* Inputs:             User buffer pointer
*
* Output:             0 - Success, -EFAULT - memory copy problem
*
* ***********************************************************************
*/
static int aipsocdkl_ioctl_GetKLStatus( char *pBuffer )
{
    static unsigned short rc  = 0;
    static unsigned char status;
    GET_KL_STATUS_PARMS gksp = {0};

    DPRINTK( "Get_KL_Status\n");

    /*
    * Read the KeyLock status
    */
    outb(KL_STATUS_INDEX, INDEX_PORT_ADDRESS);
    status = inb(DATA_PORT_ADDRESS);
    status = status & 0x07;

    DPRINTK( "Status of KeyLock is = 0x%02x\n",status );

    gksp.CompletionCode = RPDONE;
    gksp.Slot           = 0;
    gksp.KLStatus       = status;

    if ((rc = copy_to_user(pBuffer, &gksp, (unsigned long)sizeof(GET_KL_STATUS_PARMS))))
    {
        EPRINTK( "Error while copying to user space!\n");
        return -EFAULT;
    }

    return SUCCESS;
}



/************************************************************************
*
* Function Name:      aipsocdkl_ioctl_OpenCD
*
* Purpose:            Open the Cash Drawer
*
* Description:        This fn. writes a 1 to bit 0 of the port at index 0x29
*                     which triggers the ASIC to fire the Cash Drawer pulse
*
* Inputs:             User buffer pointer
*
* Output:             0 - Success, -EFAULT - memory copy problem
*
* ***********************************************************************
*/
static int aipsocdkl_ioctl_OpenCD( char *pBuffer )
{
    static unsigned short rc  = 0;
    OPEN_CD_PARMS gcdp = {0};

    DPRINTK( "Open_CD\n");

    /*
    * Fire Cash Drawer
    */
    outb(CD_OPEN_INDEX, INDEX_PORT_ADDRESS);
    outb(CD_OPEN, DATA_PORT_ADDRESS);

    gcdp.CompletionCode = RPDONE;
    gcdp.Slot           = 0;
    gcdp.WhichCD        = 1;

    if ((rc = copy_to_user(pBuffer, &gcdp, (unsigned long)sizeof(OPEN_CD_PARMS))))
    {
        WPRINTK( "Error while copying to user space!\n");
        return -EFAULT;
    }

    return SUCCESS;
}



/************************************************************************
*
* Function Name:      aipsocdkl_ioctl_GetCDStatus
*
* Purpose:            Get Cash Drawer status
*
* Description:        This fn. reads the port at index 0x30 and
*
* Inputs:             User buffer pointer
*
* Output:             0 - Success, -EFAULT - memory copy problem
*
* ***********************************************************************
*/
static int aipsocdkl_ioctl_GetCDStatus( char *pBuffer )
{
    static unsigned short rc  = 0;
    static unsigned char status;
    GET_CD_STATUS_PARMS gcsp = {0};

    DPRINTK( "Get_CD_Status\n");

    /*
    * Get CD status and mask in bits 2..1, /Open .. /Connected,
    */
    outb(CD_STATUS_INDEX, INDEX_PORT_ADDRESS);
    status = inb(DATA_PORT_ADDRESS);
    status = status & 0x06;

    DPRINTK( "Status of CD is = %x\n", status);

    gcsp.CompletionCode = RPDONE;
    gcsp.Slot           = 0;
    gcsp.CDStatus       = status;

    if ((rc = copy_to_user(pBuffer, &gcsp,(unsigned long)sizeof(GET_CD_STATUS_PARMS))))
    {
        WPRINTK( "Error while copying to user space!\n");
        return -EFAULT;
    }

    return SUCCESS;
}



/************************************************************************
*
* Function Name:      aipsocdkl_ioctl_GetSlotInfo
*
* Purpose:            Get 'Adapter' slot info
*
* Description:        Called by client to read Adapter slot info
*
* Inputs:             User buffer pointer
*
* Output:             0 - Success, -EFAULT - memory copy problem
*
* Notes:              Hard Coded values
*
* ***********************************************************************
*/
static int aipsocdkl_ioctl_GetSlotInfo( char *pBuffer )
{
    static unsigned short rc  = 0;
    GET_SLOT_INFO_PARMS gsip = {0};

    DPRINTK( "GetSlotInfo\n");

    gsip.SlotDrawers[1]   = 1;
	gsip.AdapterID[1]     = 4614;
	if (machine_type == IBM_4613) {
		gsip.AdapterID[1] = 4613;
	}
    gsip.SlotNVRAM[1]     = 0;
    gsip.CompletionCode   = RPDONE;

    if ((rc = copy_to_user( pBuffer, &gsip, sizeof(GET_SLOT_INFO_PARMS))))
    {
        EPRINTK( "Error while copying to user space!\n");
        return -EFAULT;
    }

    return SUCCESS;
}


module_init(aipsocdkl_init);
module_exit(aipsocdkl_exit);

